#!/usr/bin/env python
# -*- coding: utf-8 -*-

# Copyright (c) 2016 IAS / CNRS / Univ. Paris-Sud
# LGPL License - see attached LICENSE file
# Author: Alexandre Beelen <alexandre.beelen@ias.u-psud.fr>

"""Series of helper function to deal with healpix maps"""

from __future__ import print_function, division

from itertools import repeat, product

import logging
import numpy as np
import healpy as hp

from astropy.io import fits
from astropy.wcs import WCS, utils as wcs_utils
from astropy.table import Table, Column
from astropy.coordinates import SkyCoord, UnitSphericalRepresentation
from astropy import units as u

from .wcs_helper import equiv_celestial, build_wcs, rot_frame
from .wcs_helper import build_wcs_profile, build_wcs_cube
from .wcs_helper import DEFAULT_SHAPE_OUT

from .decorator import _hpmap

logging.basicConfig(format='%(asctime)s -- %(levelname)s: %(message)s', level=logging.DEBUG)

__all__ = ['hp_is_nest', 'hp_celestial', 'hp_to_wcs', 'hp_to_wcs_ipx',
           'hp_project', 'gen_hpmap', 'build_hpmap', 'hpmap_key',
           'wcs_to_profile', 'hp_to_profile', 'hp_profile', 'hp_stack',
           'hp_to_aperture', 'hp_photometry']


def hp_celestial(hp_header):
    """Retrieve the celestial system used in healpix maps. From Healpix documentation this can have 3 forms :

    - 'EQ', 'C' or 'Q' : Celestial2000 = eQuatorial,
    - 'G' : Galactic
    - 'E' : Ecliptic,

    only Celestial and Galactic are supported right now as the Ecliptic coordinate system was just recently pulled to astropy

    Similar to :class:`~astropty.wcs.utils.wcs_to_celestial_frame` but for header from healpix maps

    Parameters
    ----------
    hp_header : :class:`~astropy.io.fits.header.Header`
        the header of the healpix map

    Returns
    -------
    frame : :class:`~astropy.coordinates.baseframe.BaseCoordinateFrame` subclass instance
        An instance of a :class:`~astropy.coordinates.baseframe.BaseCoordinateFrame`
        subclass instance that best matches the specified WCS.

    """
    coordsys = hp_header.get('COORDSYS')

    if coordsys:
        return equiv_celestial(coordsys)
    else:
        raise ValueError("No COORDSYS in header")


def hp_is_nest(hp_header):
    """Return True if the healpix header is in nested

    Parameters
    ----------
    hp_header : :class:`~astropy.io.fits.header.Header`
        the header

    Returns
    -------
    boolean :
        True if the header is nested
    """

    ordering = hp_header.get('ORDERING', None)

    if not ordering:
        raise ValueError("Np ordering in healpix header")

    if ordering.lower() in ['nested', 'nest']:
        return True
    elif ordering.lower() == 'ring':
        return False
    else:
        raise ValueError("Unknown ordering in healpix header")


def rotate_frame(alon, alat, hp_header, wcs):
    """Change frame if needed

    Parameters
    ----------
    alon, alat : array_like
        longitudes and latitudes describe by the
    hp_header : :class:`astropy.fits.header.Header`
        corresponding header
    wcs : :class:`~astropy.wcs.WCS`
        An corresponding wcs object

    Returns
    -------
    tuple of array_like
        longitude and latitude arrays rotated if neeed
    """

    # Determine if we need a rotation for different coordinate systems
    frame_in = hp_celestial(hp_header)
    frame_out = wcs_utils.wcs_to_celestial_frame(wcs)

    if not frame_in.is_equivalent_frame(frame_out):
        logging.debug('... converting coordinate system')
        coords = frame_out.realize_frame(
            UnitSphericalRepresentation(alon * u.deg, alat * u.deg)).transform_to(frame_in)
        alon = coords.data.lon.deg
        alat = coords.data.lat.deg

    return alon, alat


def wcs_to_profile(hdu, wcs, shape_out=DEFAULT_SHAPE_OUT[0]):
    """Centered profile from 2D map

    Parameters
    ----------
    hdu : :class:`astropy.fits.ImageHDU`
        hdu containing the 2D array and corresponding header, the profile will be made from the CRVAL position
    wcs : :class:`astropy.wcs.WCS`
        wcs object to describe the radius of the profile
    shape_out : int
        shape of the output profile

    Returns
    -------
    :class:`astropy.fits.ImageHDU`
        1D hdu image containing the profile and the corresponding header
    """

    frame = wcs_utils.wcs_to_celestial_frame(WCS(hdu.header)).name
    coord = SkyCoord(hdu.header['CRVAL1'], hdu.header['CRVAL2'], unit="deg", frame=frame)

    shape = hdu.shape
    yy, xx = np.indices(shape)
    lon_arr, lat_arr = WCS(hdu.header).wcs_pix2world(xx, yy, 0)

    coords = SkyCoord(lon_arr, lat_arr, unit="deg", frame=frame)
    dist = coords.separation(coord).to(u.deg).value

    r_edge = wcs.all_pix2world(np.arange(shape_out + 1) - 0.5, 0)[0]

    hist, bin_edges = np.histogram(dist, bins=r_edge, weights=hdu.data)
    hist_d, bin_edges_d = np.histogram(dist, bins=r_edge)

    with np.errstate(invalid='ignore'):
        hist /= hist_d

    return hist


@_hpmap
def hp_to_profile(hp_hdu, wcs, coord, shape_out=DEFAULT_SHAPE_OUT[0], std=False):
    """Extract radial profile from healpix map

    Parameters
    ----------
    hp_hdu : `:class:astropy.io.fits.ImageHDU`
        a pseudo ImageHDU with the healpix map and the associated header
    wcs : :class:`astropy.wcs.WCS`
        wcs object to describe the radius of the profile
    coord : :class:`astropy.coordinate.SkyCoord`
        the sky coordinate of the center of the profile
    shape_out : int
        shape of the output profile
    std : bool
        return the standard deviation

    Returns
    -------
    :class:`astropy.fits.ImageHDU`
        1D hdu image containing the profile and the corresponding header,
        optionnaly a second ImageHDU containing the standard deviation
    """

    coord = rot_frame(coord, hp_hdu.header['COORDSYS'])

    r_edge = wcs.all_pix2world(np.arange(shape_out + 1) - 0.5, 0)[0]

    nside = hp_hdu.header['NSIDE']
    nest = hp_is_nest(hp_hdu.header)

    # TODO : can we have a product here...
    vec = hp.rotator.dir2vec(coord.data.lon.deg, coord.data.lat.deg, lonlat=True)

    # TODO: Another possibility would be to use the outer ring, compute all distances and use np.histogram
    i_pix = [hp.query_disc(nside, vec, np.radians(radius), nest=nest, inclusive=False) for radius in r_edge]

    # Only keep the pixels why are only present in the outer
    # radius
    a_pix = []
    for in_pix, out_pix in zip(i_pix[:-1], i_pix[1:]):
        a_pix.append(np.setxor1d(in_pix, out_pix))

    profile = np.asarray([np.mean(hp_hdu.data[pix]) for pix in a_pix])
    if not std:
        return profile
    else:
        std_profile = np.asarray([np.std(hp_hdu.data[pix]) for pix in a_pix])
        return profile, std_profile


@_hpmap
def hp_to_wcs(hp_hdu, wcs, shape_out=DEFAULT_SHAPE_OUT, order=0):
    """Project an Healpix map on a wcs header, using nearest neighbors.

    Parameters
    ----------
    hp_hdu : `:class:astropy.io.fits.ImageHDU`
        a pseudo ImageHDU with the healpix map and the associated header
    wcs : :class:`astropy.wcs.WCS`
        wcs object to project with
    shape_out : tuple
        shape of the output map (n_y, n_x)
    order : int (0|1)
        order of the interpolation 0: nearest-neighbor, 1: bi-linear interpolation

    Returns
    -------
    array_like
        the projected map in a 2D array of shape shape_out
    """

    # Array of pixels centers -- from the astropy.wcs documentation :
    #
    # Here, *origin* is the coordinate in the upper left corner of the
    # image.In FITS and Fortran standards, this is 1.  In Numpy and C
    # standards this is 0.
    #
    # This as we are using the C convention, the center of the first pixel is 0, 0
    y_tab, x_tab = np.indices(shape_out)

    alon, alat = wcs.wcs_pix2world(x_tab, y_tab, 0)
    # mask for pixels lying outside of projected area
    mask = ~np.logical_or(np.isnan(alon), np.isnan(alat))
    proj_map = np.ma.array(np.zeros(shape_out), mask=~mask, fill_value=np.nan)

    # Determine if we need a rotation for different coordinate systems
    alon, alat = rotate_frame(alon, alat, hp_hdu.header, wcs)

    # Healpix conventions...
    phi = np.radians(alon)
    theta = np.radians(90 - alat)

    if order == 0:
        ipix = hp.ang2pix(hp.npix2nside(len(hp_hdu.data)),
                          theta[mask], phi[mask],
                          nest=hp_is_nest(hp_hdu.header))
        proj_map[mask] = hp_hdu.data[ipix]
    elif order == 1:
        proj_map[mask] = hp.get_interp_val(
            hp_hdu.data, theta[mask], phi[mask], nest=hp_is_nest(hp_hdu.header))
    else:
        raise ValueError("Unsupported order for the interpolation")

    return proj_map.filled()


def hp_to_wcs_ipx(hp_header, wcs, shape_out=DEFAULT_SHAPE_OUT):
    """Return the indexes of pixels of a given wcs and shape_out,
    within a nside healpix map.

    Parameters
    ----------
    hp_header : :class:`astropy.fits.header.Header`
        header of the healpix map, should contain nside and coordsys and ordering
    wcs : :class:`astropy.wcs.WCS`
        wcs object to project with
    shape_out : tuple
        shape of the output map (n_y, n_x)

    Returns
    -------
    2D array_like
        mask for the given map
    array_like
        corresponding pixel indexes

    Notes
    -----
    The map could then easily be constructed using

    proj_map = np.ma.array(np.zeros(shape_out), mask=~mask, fill_value=np.nan)
    proj_map[mask] = healpix_map[ipix]
    """

    # Array of pixels centers -- from the astropy.wcs documentation :
    #
    # Here, *origin* is the coordinate in the upper left corner of the
    # image.In FITS and Fortran standards, this is 1.  In Numpy and C
    # standards this is 0.
    #
    # This as we are using the C convention, the center of the first pixel is 0, 0
    y_tab, x_tab = np.indices(shape_out)

    alon, alat = wcs.wcs_pix2world(x_tab, y_tab, 0)
    # mask for pixels lying outside of projected area
    mask = ~np.logical_or(np.isnan(alon), np.isnan(alat))

    nside = hp_header['NSIDE']

    # Determine if we need a rotation for different coordinate systems
    alon, alat = rotate_frame(alon, alat, hp_header, wcs)

    # Healpix conventions...
    phi = np.radians(alon)
    theta = np.radians(90 - alat)

    ipix = hp.ang2pix(nside, theta[mask], phi[mask], nest=hp_is_nest(hp_header))

    return mask, ipix


@_hpmap
def hp_project(hp_hdu, coord, pixsize=0.01, shape_out=DEFAULT_SHAPE_OUT, order=0, projection=('GALACTIC', 'TAN')):
    """Project an healpix map at a single given position

    Parameters
    ----------
    hp_hdu : `:class:astropy.io.fits.ImageHDU`
        a pseudo ImageHDU with the healpix map and the associated header to be projected
    coord : :class:`astropy.coordinate.SkyCoord`
        the sky coordinate of the center of the projection
    pixsize : float
        size of the pixel (in degree)
    shape_out : tuple
        shape of the output map (n_y, n_x)
    order : int (0|1)
        order of the interpolation 0: nearest-neighbor, 1: bi-linear interpolation
    projection : tuple of str
        the coordinate ('GALACTIC', 'EQUATORIAL') and projection ('TAN', 'SIN', 'GSL', ...) system

    Returns
    -------
    :class:`astropy.io.fits.PrimaryHDU`
        containing the array and the corresponding header
    """

    proj_sys, proj_type = projection

    wcs = build_wcs(coord, pixsize, shape_out=shape_out, proj_sys=proj_sys, proj_type=proj_type)
    proj_map = hp_to_wcs(hp_hdu, wcs, shape_out=shape_out, order=order)

    return fits.ImageHDU(proj_map, wcs.to_header(relax=0x20000))


@_hpmap
def hp_stack(hp_hdu, coords, pixsize=0.01, shape_out=DEFAULT_SHAPE_OUT, order=0, projection=('GALACTIC', 'TAN'), keep=False):
    """Perform stacking on an healpix map

    Parameters
    ----------
    hp_hdu : `:class:astropy.io.fits.ImageHDU`
        a pseudo ImageHDU with the healpix map and the associated header to be projected
    coords : list of :class:`astropy.coordinate.SkyCoord`
        list of sky coordinates for the center of the cropped maps
    pixsize : float
        size of the pixel (in degree)
    shape_out : tuple
        shape of the output map (n_y, n_x)
    order : int (0|1)
        order of the interpolation 0: nearest-neighbor, 1: bi-linear interpolation
    projection : tuple of str
        the coordinate ('GALACTIC', 'EQUATORIAL') and projection ('TAN', 'SIN', 'GSL', ...) system
    keep : boolean (default False)
        return all the cropped maps as a 3D cube instead of one stack map

    Returns
    -------
    `:class:~fits.ImageHDU`
        hdu containing the stack image or cube and corresponding header
    """

    proj_sys, proj_type = projection

    if coords.isscalar:
        coords = SkyCoord([coords.frame])

    coords = rot_frame(coords, proj_sys)

    if isinstance(pixsize, (int, float)):
        pixsize = repeat(pixsize, coords.shape[0])
    else:
        pixsize = iter(pixsize)

    w = build_wcs_cube(coords[0], 0, pixsize=next(pixsize), shape_out=shape_out,
                       proj_sys=proj_sys, proj_type=proj_type)

    # First image outside of the loop
    _w = w.dropaxis(2)
    stacks = hp_to_wcs(hp_hdu, _w, shape_out)

    if keep:
        # return 3D array
        stacks = np.ma.resize(stacks, coords.shape + stacks.shape)

    for index, (coord, pix) in enumerate(zip(coords[1:], pixsize), 1):
        _w.wcs.crval = (coord.data.lon.deg, coord.data.lat.deg)
        _w.wcs.cdelt = [-pix, pix]

        if keep:
            stacks[index, :, :] = hp_to_wcs(hp_hdu, _w, shape_out)
        else:
            stacks += hp_to_wcs(hp_hdu, _w, shape_out)

    # Take the mean
    if not keep:
        stacks /= coords.shape[0]
        w = w.dropaxis(2)

    if coords.shape[0] > 1:
        w.wcs.crval[0:2] = [0, 0]

    return fits.ImageHDU(stacks, w.to_header(relax=0x20000))


@_hpmap
def hp_profile(hp_hdu, coord, pixsize=0.01, npix=512):
    """Project an healpix map at a single given position

    Parameters
    ----------
    hp_hdu : `:class:astropy.io.fits.ImageHDU`
        a pseudo ImageHDU with the healpix map and the associated header to be projected
    coord : :class:`astropy.coordinate.SkyCoord`
        the sky coordinate of the center of the projection
    pixsize : float
        size of the pixel (in degree)
    npix : int
        number of pixels in the final map, the reference pixel will be at the center

    Returns
    -------
    :class:`astropy.io.fits.PrimaryHDU`
        containing the array and the corresponding header
    """

    wcs = build_wcs_profile(pixsize)
    profile = hp_to_profile(hp_hdu, wcs, coord, shape_out=npix, std=False)

    return fits.ImageHDU(profile, wcs.to_header(relax=0x20000))


@_hpmap
def hp_to_aperture(hp_hdu, coords, apertures=None):
    """Raw aperture summation on an healpix map

    Parameters
    ----------
    hp_hdu : `:class:astropy.io.fits.ImageHDU`
        a pseudo ImageHDU with the healpix map and the associated header to be projected
    coords : list of :class:`astropy.coordinate.SkyCoord`
        the sky coordinates for the center of the apertures
    apertures : list of `:class:astropy.coordinates.Angles`
        aperture angle in which we perfom summation

    Returns
    -------
    npix, apertures : array_like
        2 arrays containing the number of pixels, and sum of the pixels
        within the aperture respectively
    """

    coords = rot_frame(coords, hp_hdu.header['COORDSYS'])

    nside = hp_hdu.header['NSIDE']
    nest = hp_is_nest(hp_hdu.header)

    vecs = hp.rotator.dir2vec(coords.data.lon.deg, coords.data.lat.deg, lonlat=True)

    if coords.isscalar or coords.shape == (1,):
        vecs = vecs[:, np.newaxis]

    pix_indexes = [hp.query_disc(nside, vec, np.radians(radius.degree), nest=nest, inclusive=False)
                   for vec, radius in product(vecs.T, apertures)]

    # TODO: have a look at regions.PixelRegion for exact boundings
    # apertures = [(len(index), np.sum(hp_hdu.data[index])) if len(index) > 0 else None for index in pix_indexes]
    npix, apertures = zip(*[(index.shape[0], np.sum(hp_hdu.data[index])) for index in pix_indexes])

    return np.reshape(npix, coords.shape + (-1,)), np.reshape(apertures, coords.shape + (-1,))


@_hpmap
def hp_photometry(hp_hdu, coords, apertures):
    """Aperture photometry on an healpix map at a single given position

    Parameters
    ----------
    hp_hdu : `:class:astropy.io.fits.ImageHDU`
        a pseudo ImageHDU with the healpix map and the associated header to be projected
    coords : :class:`astropy.coordinate.SkyCoord`
        the sky coordinates for the center of the apertures
    apertures : 3 `:class:astropy.coordinates.Angles`
        3 floats defining the aperture radius and inner/outer annulus radii

    Returns
    -------
    :class:`astropy.io.fits.BinaryHDU`
        table containing the photometry
    """

    if isinstance(coords, list):
        coords = SkyCoord(coords)
    elif coords.isscalar:
        coords = SkyCoord([coords])

    npix, S = hp_to_aperture(hp_hdu, coords, apertures)

    S_aper, S_inner, S_outer = S.T
    n_aper, n_inner, n_outer = npix.T

    # Perfom photometry
    mean_background = (S_outer - S_inner) / (n_outer - n_inner)
    mean_brigthness = S_aper / n_aper - mean_background

    result = Table()
    result['brigthness'] = Column(mean_brigthness, unit=hp_hdu.header['UNIT'])
    result['background'] = Column(mean_background, unit=hp_hdu.header['UNIT'])
    result['n_pix'] = Column(n_aper)

    return result


def gen_hpmap(maps):
    """Generator function for large maps and low memory system

    Parameters
    ----------
    maps : list
        A list of Nmap tuples with either:
            * (filename, path_to_localfilename, healpix header)
            * (filename, healpix vector, healpix header)

    Returns
    -------
    tuple
        Return a tuple (filename, healpix map, healpix header) corresponding to the inputed list
    """
    for filename, i_map, i_header in maps:
        if isinstance(i_map, str):
            i_map = hp.read_map(i_map, verbose=False, nest=None)
            i_map = hp.ma(i_map)
        yield filename, fits.ImageHDU(i_map, fits.Header(i_header))


def build_hpmap(filenames, low_mem=True):
    """From a filename list, build a tuple usable with gen_hmap()

    Parameters
    ----------
    filenames: list
        A list of Nmap filenames of healpix maps
    low_mem : bool
        On low memory system, do not read the maps themselves (default: only header)

    Returns
    -------
    tuple list
        A list of tuple which can be used by gen_hpmap    """

    hp_maps = []
    for filename in filenames:
        hp_header = fits.getheader(filename, 1)
        if low_mem is True:
            hp_map = filename
        else:
            hp_map = hp.read_map(filename, verbose=False, nest=None)
            hp_map = hp.ma(hp_map)
        hp_maps.append((filename, hp_map, hp_header))
    return hp_maps


def hpmap_key(hp_map):
    """Generate an key from the hp_map tuple to sort the hp_maps by map
    properties

    Parameters
    ----------
    hp_map: tuple
        A tuple from (build|gen)_hpmap : (filename, healpix map, healpix header)

    Returns
    -------
    str
        A string with the map properties
    """
    i_header = hp_map[2]

    return "%s_%s_%s" % (i_header['NSIDE'], i_header['ORDERING'], hp_celestial(i_header).name)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# Copyright (c) 2016 IAS / CNRS / Univ. Paris-Sud
# LGPL License - see attached LICENSE file
# Author: Alexandre Beelen <alexandre.beelen@ias.u-psud.fr>

"""parse module, to be used by :func:`cutsky.main`, mainly here to reduce Cyclomatic complexcity"""

from __future__ import print_function, division

import os
import sys
import argparse
import logging

from .wcs_helper import VALID_PROJ

try:  # pragma: py3
    from configparser import ConfigParser, ExtendedInterpolation
    PATTERN = None
except ImportError:  # pragma: py2
    import re
    from ConfigParser import ConfigParser
    PATTERN = re.compile(r"\$\{(.*?)\}")

try:  # pragma: py3
    FileNotFoundError
except NameError:  # pragma: py2
    FileNotFoundError = IOError

LOGGER = logging.getLogger('django')

DEFAULT_NPIX = 256
DEFAULT_PIXSIZE = 1
DEFAULT_COORDFRAME = 'galactic'
DEFAULT_CTYPE = 'TAN'

__all__ = ['parse_args', 'parse_config', 'combine_args_config', 'ini_main']


def parse_args(args):
    """Parse arguments from the command line"""

    parser = argparse.ArgumentParser(
        description="Reproject the spherical sky onto a plane.")
    parser.add_argument('lon', type=float,
                        help='longitude of the projection [deg]')
    parser.add_argument('lat', type=float,
                        help='latitude of the projection [deg]')

    # Removed the default argument from here otherwise the config file
    # wont be used

    group = parser.add_mutually_exclusive_group()
    group.add_argument('--npix', type=int,
                       help='number of pixels (default 256)')
    group.add_argument('--radius', type=float,
                       help='radius of the requested region [deg] ')

    parser.add_argument('--pixsize', type=float,
                        help='pixel size [arcmin] (default 1)')

    parser.add_argument('--coordframe', required=False,
                        help='coordinate frame of the lon. and \
                        lat. of the projection and the projected map \
                        (default: galactic)',
                        choices=['galactic', 'fk5'])
    parser.add_argument('--ctype', required=False,
                        help='any projection code supported by wcslib\
                         (default:TAN)',
                        choices=VALID_PROJ)

    input_map = parser.add_argument_group(
        'input maps', description="one of the two options must be present")

    input_map.add_argument('--mapfilenames', nargs='+', required=False,
                           help='absolute path to the healpix maps')
    input_map.add_argument('--conf', required=False,
                           help='absolute path to a config file')

    out = parser.add_argument_group('output')
    out.add_argument('--fits', action='store_true', help='output fits file')
    out.add_argument('--png', action='store_true',
                     help='output png file (Default: True if nothing else)')
    out.add_argument(
        '--votable', action='store_true', help='output votable file')
    out.add_argument('--outdir', required=False,
                     help='output directory (default:".")')

    general = parser.add_argument_group('general')
    verb = general.add_mutually_exclusive_group()
    verb.add_argument(
        '-v', '--verbose', action='store_true', help='verbose mode')
    verb.add_argument('-q', '--quiet', action='store_true', help='quiet mode')

    # Do the actual parsing
    parsed_args = parser.parse_args(args)

    # Put the list of filenames into the same structure as the config
    # file, we are loosing the doContour keyword but...
    if parsed_args.mapfilenames:
        parsed_args.maps = [(filename,
                             dict([('legend', os.path.splitext(os.path.basename(filename))[0]), ('doContour', False)]))
                            for filename in
                            parsed_args.mapfilenames]
    else:
        parsed_args.maps = None

    if parsed_args.verbose:
        parsed_args.verbosity = logging.DEBUG
    elif parsed_args.quiet:
        parsed_args.verbosity = logging.ERROR
    else:
        parsed_args.verbosity = None

    return parsed_args


def find_config(conffile=None):
    """Read the configuration file."""

    if PATTERN:  # pragma: py2
        config = ConfigParser()

    else:  # pragma: py3
        config = ConfigParser(interpolation=ExtendedInterpolation())

    # Look for cutsky.cfg at several locations
    conffiles = [os.path.join(directory, 'cutsky.cfg') for directory
                 in [os.curdir, os.path.join(os.path.expanduser("~"), '.config/cutsky')]]

    # If specifically ask for a config file, then put it at the
    # beginning ...
    if conffile:
        conffiles.insert(0, conffile)

    found = config.read(conffiles)

    if not found:
        raise FileNotFoundError

    return config


def parse_config_basic(config):
    """Parse basic options from the configuration file."""

    # Basic options, same as the option on the command line
    options = {}
    if config.has_option('cutsky', 'npix'):
        options['npix'] = config.getint('cutsky', 'npix')
    if config.has_option('cutsky', 'pixsize'):
        options['pixsize'] = config.getfloat('cutsky', 'pixsize')

    for key in ['coordframe', 'ctype', 'outdir']:
        if config.has_option('cutsky', key):
            options[key] = config.get('cutsky', key)

    return options


def parse_config_basic_output(config, options):
    """Parse basic output options from the configuration file."""

    for key in ['fits', 'png', 'votable']:
        if config.has_option('cutsky', key) and config.get('cutsky', key):
            options[key] = True

    return options


def parse_config_verbosity(config):
    """Parse the verbosity level from the config file."""

    allowed_levels = {'verbose': logging.DEBUG,
                      'debug': logging.DEBUG,
                      'quiet': logging.ERROR,
                      'error': logging.ERROR,
                      'info': logging.INFO}

    verbosity = None

    if config.has_option('cutsky', 'verbosity'):
        level = str(config.get('cutsky', 'verbosity')).lower()

        if level in allowed_levels.keys():
            verbosity = allowed_levels[level]
        elif level.isdigit():
            verbosity = int(level)

    return verbosity


def parse_filename(config, section):
    """Parse filename from a config file py2/py3 compatible"""

    filename = config.get(section, 'filename')
    if PATTERN:  # pragma: v2
        def rep_key(mesg):
            """re.sub-pattern to find an extendedInterpolation in py2"""
            section, key = re.split(':', mesg.group(1))
            return config.get(section, key)
        filename = PATTERN.sub(rep_key, filename)

    return filename


def parse_config_map_opt(config, section):
    """Parse map option in a configuration file."""

    opt = {'legend': section}

    if config.has_option(section, 'doContour'):
        opt['doContour'] = config.getboolean(section, 'doContour')
    else:
        opt['doContour'] = False

    return opt


def parse_config_select_maps(config, sections):
    """Retrieve the section of the maps to project from a configuration file."""

    good_sections = []
    for key in sections:
        if not config.has_option(key, 'doCut') or config.getboolean(key, 'doCut'):
            good_sections.append(key)

    return good_sections


def parse_config_maps(config):
    """Parse the maps options from the configuration file."""

    # Map list, only get the one which will be projected
    # Also check if contours are requested
    maps_to_cut = []

    map_sections = [key for key in config.sections() if key != 'cutsky' and config.has_option(key, 'filename')]

    map_sections = parse_config_select_maps(config, map_sections)

    for section in map_sections:

        filename = parse_filename(config, section)

        opt = parse_config_map_opt(config, section)

        maps_to_cut.append((filename, opt))

    return maps_to_cut


def parse_config(conffile=None):
    """Parse options from a configuration file."""

    config = find_config(conffile)

    options = parse_config_basic(config)

    options = parse_config_basic_output(config, options)

    options['verbosity'] = parse_config_verbosity(config)
    options['maps'] = parse_config_maps(config)

    return options


def combine_args_config(args, config):
    """
    Combine the different sources of arguments (command line,
    configfile or default arguments

    Notes
    -----
    The `radius` arguments have priority and `npix` will be computed
    based on `pixelsize`
    """

    arguments = [('pixsize', DEFAULT_PIXSIZE),
                 ('coordframe', DEFAULT_COORDFRAME),
                 ('ctype', DEFAULT_CTYPE),
                 ('verbosity', logging.INFO),
                 ('maps', None),
                 ('fits', False),
                 ('png', False),
                 ('votable', False),
                 ('outdir', '.'), ]

    combined_args = vars(args)
    for key, default_value in arguments:
        combined_args[key] = combined_args[key] or config.get(key) or default_value

    LOGGER.setLevel(combined_args['verbosity'])

    # setting the radius will define the number of pixels
    if args.radius:
        args.npix = int(args.radius / (float(combined_args['pixsize']) / 60))

    combined_args['npix'] = args.npix or config.get('npix') or DEFAULT_NPIX

    if not (combined_args['fits'] or combined_args['png'] or combined_args['votable']):
        combined_args['png'] = True

    return combined_args


def ini_main(argv):
    """Initialize the main function"""

    if argv is None:  # pragma: no cover
        argv = sys.argv[1:]

    try:
        args = parse_args(argv)
    except SystemExit:  # pragma: no cover
        sys.exit()

    try:
        config = parse_config(args.conf)
    except FileNotFoundError:
        config = {}

    return combine_args_config(args, config)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# Copyright (c) 2016 IAS / CNRS / Univ. Paris-Sud
# LGPL License - see attached LICENSE file
# Author: Alexandre Beelen <alexandre.beelen@ias.u-psud.fr>

"""cutsky module, mainly use :mod:`hpproj.hp_helper` functions"""

import warnings
import logging

import os
import sys
from base64 import b64encode
from itertools import groupby

try:  # pragma: py3
    from io import BytesIO
except ImportError:  # pragma: py2
    from cStringIO import StringIO as BytesIO

import numpy as np

# try: # prama: no cover
# from wcsaxes import WCS # (deprecated)
# except ImportError: # pragma: no cover
from astropy.wcs import WCS
from astropy import units as u
from astropy.io import fits
from astropy.table import Table
from astropy.coordinates import SkyCoord

from photutils import CircularAperture
from photutils import aperture_photometry

from .wcs_helper import VALID_EQUATORIAL
from .wcs_helper import equiv_celestial, build_wcs
from .hp_helper import build_hpmap, gen_hpmap
from .hp_helper import hp_to_wcs_ipx, hpmap_key

from .parse import ini_main
from .parse import DEFAULT_NPIX, DEFAULT_PIXSIZE, DEFAULT_COORDFRAME, DEFAULT_CTYPE

import matplotlib.pyplot as plt

try:  # pragma: py3
    FileNotFoundError
except NameError:  # pragma: py2
    FileNotFoundError = IOError

LOGGER = logging.getLogger('django')

DEFAULT_PATCH = [256, 1.]


class CutSky(object):

    """
    Container for Healpix maps and cut_* methods

    ...

    Attributes
    ----------
    npix : int
        the number of pixels for the square maps
    pixsize : float
        the size of the pixels [arcmin]
    ctype : str
        a valid projection type (default : TAN)
    maps : dictonnary
        a grouped dictionnary of gen_hpmap tuples (filename, map, header) (see :func:~init)


    """

    def __init__(self, maps=None, npix=DEFAULT_NPIX, pixsize=DEFAULT_PIXSIZE, ctype=DEFAULT_CTYPE, low_mem=True):
        """Initialization of a CutSky class

        Parameters
        ----------
        maps : list of tuple or dictionary
            list of tuple (filename, {opt}) where filename is the full
            path to the healpix map and {opt} is a dictionnary with
            "optional" header for the file (see Notes).
        npix : int
            the number of pixels for the square maps
        pixsize : float
            the size of the pixels [arcmin]
        ctype : str
            a valid projection type (default : TAN)
        low_mem: bool
            a boolean deciding if all the maps are loaded in memory or
            not (default: True, the maps are not loaded on init)

        Notes
        -----
        The list of dictionnary describe the maps to be projected :
        ```
        [(full_filename_to_healpix_map.fits, {'legend': legend,
                                              'doContour': True}), # optionnal
          ... ]
        ```
        The {opt} dictionnary MUST containt, at least, the key
        'legend' which will be used to uniquely identify the cutted
        map, other possible keys 'doContour' with a boolean value to
        contour the map latter one

        Alternatively one can use the old interface as a dictionnary :
        ```
        {legend: {'filename': full_filename_to_healpix_map.fits,
                    'doContour': True }, # optionnal
         ... }
        ```
        """

        if maps is None:
            raise FileNotFoundError("No healpix map to project")

        if isinstance(maps, dict):
            maps = to_new_maps(maps)

        # Define the basic parameters for the output maps
        self.npix = npix
        self.pixsize = pixsize
        self.ctype = ctype

        # Build an hp_map list and extend the headers of the maps with
        # the optionnal values
        filenames, opts = zip(*maps)
        hp_map = build_hpmap(filenames, low_mem=low_mem)
        for (filename, i_map, i_header), opt in zip(hp_map, opts):
            # Insure that we do have a doContour key internally
            if 'doContour' not in opt.keys():
                opt['doContour'] = False
            i_header.extend(opt)

        # group them by map properties for efficiencies reasons
        hp_map.sort(key=hpmap_key)
        self.maps = hp_map

        # Save intermediate results
        self.maps_selection = None
        self.cuts = None
        self.lonlat = None
        self.coordframe = DEFAULT_COORDFRAME

    def cut_fits(self, lonlat=None, coordframe=DEFAULT_COORDFRAME, maps_selection=None):
        """Efficiently cut the healpix maps and return cutted fits file with proper header

        Parameters
        ----------
        lonlat : array of 2 floats
            the longitude and latitude of the center of projection [deg]
        coordframe : str
            the coordinate frame used for the position AND the projection
        maps_selection : list
            optionnal list of the 'legend' or filename of the map to
            select a sub-sample of them.

        Returns
        -------
        list of dictionnaries
            the dictionnary has 2 keys :
            * 'legend' (the opts{'legend'} see __init())
            * 'fits' an :class:`~astropy.io.fits.ImageHDU`
        """

        assert len(lonlat) == 2, "lonlat must have 2 elements"

        self.maps_selection = maps_selection

        # Center of projection
        coord_in = SkyCoord(
            lonlat[0], lonlat[1], unit=u.deg, frame=equiv_celestial(coordframe))

        # Build the target WCS header
        wcs = build_wcs(coord_in, pixsize=self.pixsize / 60.,
                        shape_out=(self.npix, self.npix), proj_sys=coordframe, proj_type=self.ctype)

        cuts = []
        for group, maps in groupby(self.maps, key=hpmap_key):
            LOGGER.info('projecting ' + group)

            # Construct a basic healpix header from the group key, this
            # will be commun for all the maps in this group
            nside, ordering, coordsys = group.split('_')
            hp_header = {'NSIDE': int(nside),
                         'ORDERING': ordering,
                         'COORDSYS': coordsys}

            # Extract mask & pixels, common for all healpix maps of this
            # group
            mask, ipix = hp_to_wcs_ipx(hp_header, wcs, shape_out=(self.npix, self.npix))

            # Set up the figure, common for all healpix maps of this group
            LOGGER.info('cutting maps')

            # Now the actual healpix map reading and projection
            for filename, i_hdu in self._to_process(gen_hpmap(maps)):
                legend = i_hdu.header['legend']
                patch = np.ma.array(
                    np.zeros((self.npix, self.npix)), mask=~mask, fill_value=np.nan)
                patch[mask] = i_hdu.data[ipix]
                header = wcs.to_header()
                header.append(('filename', filename))
                header.append(('legend', legend))
                header.append(('doContour', i_hdu.header['doContour']))

                cuts.append({'legend': legend,
                             'fits': fits.ImageHDU(patch, header)})
        self.lonlat = lonlat
        self.coordframe = coordframe
        self.cuts = cuts

        return cuts

    def _to_process(self, gen_hpmap_iter):
        """Iterator to filter the gen_hpmap iterator depending on map selection"""

        for filename, i_hdu in gen_hpmap_iter:
            legend = i_hdu.header['legend']
            if not self.maps_selection or (legend in self.maps_selection or filename in self.maps_selection):
                yield filename, i_hdu

    def _get_cuts(self, lonlat=None, coordframe=None, maps_selection=None):
        """Get map cuts if they are already made, or launch cut_fits

        Parameters
        ----------
        lonlat : array of 2 floats
            the longitude and latitude of the center of projection [deg]
        coordframe : str
            the coordinate frame used for the position AND the projection
        maps_selection : list
            optionnal list of the 'legend' or filename of the map to
            select a sub-sample of them.

        Returns
        -------
        list of dictionnaries
            the dictionnary has 2 keys :
            * 'legend' (the opts{'legend'} see __init())
            * 'fits' an :class:`~astropy.io.fits.ImageHDU`
        """

        if self.lonlat == lonlat and \
           self.coordframe == coordframe and  \
           self.maps_selection == maps_selection and \
           self.cuts:
            # Retrieve previously cut maps
            cuts = self.cuts
        else:
            # Or cut the maps
            cuts = self.cut_fits(
                lonlat=lonlat, coordframe=coordframe, maps_selection=maps_selection)

        return cuts

    def cut_png(self, lonlat=None, coordframe=DEFAULT_COORDFRAME, maps_selection=None):
        """Efficiently cut the healpix maps and return cutted fits file with proper header and corresponding png

        Parameters
        ----------
        lonlat : array of 2 floats
            the longitude and latitude of the center of projection [deg]
        coordframe : str
            the coordinate frame used for the position AND the projection
        maps_selection : list
            optionnal list of the 'legend' or filename of the map to
            select a sub-sample of them.

        Returns
        -------
        list of dictionnaries
            the dictionnary has 3 keys :
            * 'legend' (the opts{'legend'} see __init()),
            * 'fits' an :class:`~astropy.io.fits.ImageHDU`,
            * 'png', a b61encoded png image of the fits

        """

        cuts = self._get_cuts(lonlat, coordframe, maps_selection)

        patch = np.zeros((self.npix, self.npix))

        # Plotting
        old_backend = plt.get_backend()
        plt.switch_backend('Agg')
        fig = plt.figure()

        ax_wcs = fig.add_axes(
            [0.1, 0.1, 0.9, 0.9], projection=WCS(cuts[0]['fits']))
        proj_im = ax_wcs.imshow(patch, interpolation='none', origin='lower')
        ax_wcs.coords.grid(color='green', linestyle='solid', alpha=0.5)

        # DEFAULT_COORDFRAME is galactic
        axes = [('glon', 'red', r'$l$'),
                ('glat', 'red', r'$b$')]

        if np.str(coordframe) in VALID_EQUATORIAL:
            axes = [('ra', 'white', r'$\alpha_\mathrm{J2000}$'),
                    ('dec', 'white', r'$\delta_\mathrm{J2000}$')]

        for axis, color, label in axes:
            ax_wcs.coords[axis].set_ticks(color=color)
            ax_wcs.coords[axis].set_axislabel(label)

        for cut in cuts:

            legend = cut['legend']
            LOGGER.debug('plotting ' + legend)

            patch = cut['fits'].data
            patch_header = cut['fits'].header

            proj_im.set_data(patch)
            proj_im.set_clim(vmin=patch.min(), vmax=patch.max())

            # Insure we do have increasing values even when patch.std == 0
            levels = np.arange(2, 10) * (np.max([patch.std(), 1e-12]))
            if patch_header['doContour']:
                LOGGER.debug('contouring ' + legend)
                ax_wcs.contour(
                    patch, levels=levels, colors="white", interpolation='bicubic')

            # TODO: Retrive and use the contour in other plots

            LOGGER.debug('saving ' + legend)

            # Add the map to the cut dictionnary
            output_map = BytesIO()
            plt.savefig(output_map, bbox_inches='tight',
                        format='png', dpi=75, frameon=False)
            cut['png'] = b64encode(output_map.getvalue()).strip()
            LOGGER.debug(legend + ' done')

        plt.switch_backend(old_backend)
        return cuts

    def cut_phot(self, lonlat=None, coordframe=DEFAULT_COORDFRAME, maps_selection=None):
        """Efficiently cut the healpix maps and return cutted fits file with proper header and corresponding photometry

        Parameters
        ----------
        lonlat : array of 2 floats
            the longitude and latitude of the center of projection [deg]
        coordframe : str
            the coordinate frame used for the position AND the projection
        maps_selection : list
            optionnal list of the 'legend' or filename of the map to
            select a sub-sample of them.

        Returns
        -------
        list of dictionnaries
            the dictionnary has 3 keys :
            * 'legend' (the opts{'legend'} see __init()),
            * 'fits' an :class:`~astropy.io.fits.ImageHDU`,
            * 'phot', the corresponding photometry

        """

        cuts = self._get_cuts(lonlat, coordframe, maps_selection)

        positions = [(self.npix * 1. / 2., self.npix * 1. / 2)]
        apertures = CircularAperture(positions, r=3. / self.pixsize)

        for cut in cuts:
            legend = cut['legend']
            LOGGER.debug('phot on ' + legend)

            patch = cut['fits'].data
            cut['phot'] = aperture_photometry(
                patch - np.median(patch), apertures)

            LOGGER.debug(legend + ' done')

        return cuts

    def cut(self, cut_type, **kwargs):
        """helper function to cut into the maps

        Parameters
        ----------
        cut_type : str (fits|png|phot|votable)
            define what to cut_type
        lonlat : array of 2 floats
            the longitude and latitude of the center of projection [deg]
        coordframe : str
            the coordinate frame used for the position AND the projection
        maps_selection : list
            optionnal list of the 'legend' or filename of the map to
            select a sub-sample of them.

        Returns
        -------
        list of dictionnaries
            the dictionnary output depends on cut_type
        """

        if cut_type == 'fits':
            results = self.cut_fits(**kwargs)
        elif cut_type == 'png':
            results = self.cut_png(**kwargs)
        elif cut_type in ['phot', 'votable']:
            results = self.cut_phot(**kwargs)

        return results


def to_new_maps(maps):
    """Transform old dictionnary type healpix map list used by cutsky to
    list of tuple used by Cutsky

    Parameters
    ----------
    maps : dict
        a dictionnary with key being the legend of the image :
        ```
        {legend: {'filename': full_filename_to_healpix_map.fits,
                    'doContour': True },
         ... }
        ```

    Returns
    -------
    a list of tuple following the new convention:
    ```
    [(full_filename_to_healpix_map.fits, {'legend': legend,
                                          'doContour': True}),
     ... ]
    ```
    """

    warnings.warn("deprecated", DeprecationWarning)

    new_maps = []
    for key in iter(maps.keys()):
        filename = maps[key]['filename']
        opt = {'legend': key}
        if 'doContour' in maps[key]:
            opt['doContour'] = maps[key]['doContour']
        new_maps.append((filename, opt))

    return new_maps


def cutsky(lonlat=None, maps=None, patch=None, coordframe=DEFAULT_COORDFRAME, ctype=DEFAULT_CTYPE):
    """Old interface to cutsky -- Here mostly for compability

    Parameters
    ----------
    lonlat : array of 2 floats
        the longitude and latitude of the center of projection [deg]
    maps: a dict or a list
        either a dictionnary (old interface) or a list of tuple (new
        interface) :
        ```
        {legend: {'filename': full_filename_to_healpix_map.fits,
                  'doContour': True }, # optionnal
         ... }
         ```
         or
         ```
         [(full_filename_to_healpix_map.fits, {'legend': legend,
                                              'doContour': True}), # optionnal
         ... ]
         ```
    patch : array of [int, float]
        [int] the number of pixels and
        [float] the size of the pixel [arcmin]
    coordframe : str
        the coordinate frame used for the position AND the projection
    ctype: str
        a valid projection type (default: TAN)

    Returns
    -------
    list of dictionnaries
        the dictionnary has 4 keys :
        * 'legend' (see maps above),
        * 'fits' an :class:`~astropy.io.fits.ImageHDU`,
        * 'png', a b61encoded png image of the fits
        * 'phot', the corresponding photometry

    """

    assert lonlat is not None, "You must provide a lonlat argument"
    assert len(lonlat) == 2, "lonlat must have 2 elements"
    assert maps is not None, "You must provide an healpix map to project"

    if patch is None:
        patch = DEFAULT_PATCH

    if isinstance(maps, dict):
        maps = to_new_maps(maps)

    cut_those_maps = CutSky(
        maps=maps, npix=patch[0], pixsize=patch[1], ctype=ctype)
    result = cut_those_maps.cut_png(lonlat=lonlat, coordframe=coordframe)
    result = cut_those_maps.cut_phot(lonlat=lonlat, coordframe=coordframe)

    return result


def save_result(output, result):
    """Save the results of the main function"""

    from base64 import b64decode

    filename = os.path.join(output['outdir'], result['legend'])
    if output['fits']:
        # Works only on more recent version of astropy...
        # try:
        #     hdulist = fits.HDUList([fits.PrimaryHDU(), result['fits']])
        #     hdulist.writeto(filename + '.fits', clobber=True)
        # except NotImplementedError:
        # So for now...
        result['fits'].data = result['fits'].data.filled()
        hdulist = fits.HDUList([fits.PrimaryHDU(), result['fits']])
        hdulist.writeto(filename + '.fits', overwrite=True)

    if output['png']:
        png = open(filename + '.png', 'wb')
        png.write(b64decode(result['png']))
        png.close()

    if output['votable']:
        # Need to cast into astropy Table before writing to votable
        phot = Table(result['phot'])
        with open(filename + '.xml', 'w') as file_out:
            phot.write(file_out, format='votable')


def main(argv=None):
    """The main routine."""

    args = ini_main(argv)

    cut_those_maps = CutSky(maps=args['maps'], npix=args['npix'], pixsize=args['pixsize'], ctype=args['ctype'])
    for key in ['fits', 'png', 'votable']:
        if args[key]:
            results = cut_those_maps.cut(
                key, lonlat=[args['lon'], args['lat']], coordframe=args['coordframe'])

    if not os.path.isdir(args['outdir']):
        os.makedirs(args['outdir'])

    for result in results:
        save_result(args, result)


if __name__ == '__main__':
    main(sys.argv[1:])



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# Copyright (c) 2016 IAS / CNRS / Univ. Paris-Sud
# LGPL License - see attached LICENSE file
# Author: Alexandre Beelen <alexandre.beelen@ias.u-psud.fr>

"""Series of decorator functions"""

import numpy as np
from astropy.io import fits

__all__ = ['_hpmap', 'update_docstring']


# Decorator and decorated functions
def update_docstring(fct, skip=0, head_docstring=None, foot_docstring=None):
    """Update the docstring of a decorated function by insterting a docstring at the beginning

    Parameters
    ----------
    fct : fct
        the original function from which we extract the documentation
    skip : int
        number of line to skip in the original documentation
    head_docstring : str
        additionnal documentation to be insterted at the top
    foot_docstring : str
        additionnal documentation to be inserted at the end

    Returns
    -------
    fct
        the decorator function with the updated docstring
    """

    # Extract first line of the original function
    docstring = str(fct.__doc__).split('\n')[0]

    if head_docstring:
        docstring += head_docstring

    docstring += '\n' + '\n'.join(str(fct.__doc__).split('\n')[skip:])

    if foot_docstring:
        docstring += foot_docstring

    return docstring


def _hpmap(hphdu_func):
    """Will decorate the function taking hp_hdu function to be able to use it with hp_map, hp_header instead of an :class:`astropy.io.fits.ImageHDU` object

    Parameters
    ----------
    a function using hp_hpu :class:`astropy.io.fits.ImageHDU` as the first argument

    Returns
    -------
    The same function decorated

    Notes
    -----
    To use this decorator
    hp_to_wcs = _hpmap(hphdu_to_wcs)
    or use @_hpmap on the function declaration
    """

    def decorator(*args, **kargs):
        """Transform a function call from (hp_map, hp_header,*) to (ImageHDU, *)"""
        if isinstance(args[0], np.ndarray) and \
           (isinstance(args[1], dict) or isinstance(args[1], fits.Header)):
            hp_hdu = fits.ImageHDU(args[0], fits.Header(args[1]))
            # Ugly fix to modigy the args
            args = list(args)
            args.insert(2, hp_hdu)
            args = tuple(args[2:])
        return hphdu_func(*args, **kargs)

    decorator._hphdu = hphdu_func
    decorator.__doc__ = update_docstring(hphdu_func, skip=6, head_docstring="""

    Parameters
    ----------
    hp_hdu : :class:`astropy.io.fits.ImageHDU`
        a pseudo ImageHDU with the healpix map and the associated header

        or

    hp_map : array_like
        healpix map with corresponding...
    hp_header : :class:`astropy.fits.header.Header`
        ...header""", foot_docstring="""
    Notes
    -----
    You can access a function using only catalogs with the ._coord() method
    """)
    return decorator



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# Copyright (c) 2016 IAS / CNRS / Univ. Paris-Sud
# LGPL License - see attached LICENSE file
# Author: Alexandre Beelen <alexandre.beelen@ias.u-psud.fr>

"""Series of helper function to deal with building wcs objects"""

import logging
import numpy as np

from astropy.wcs import WCS
from astropy.coordinates import ICRS, Galactic, Angle, SkyCoord

from .decorator import update_docstring

VALID_PROJ = ['AZP', 'SZP', 'TAN', 'STG', 'SIN',
              'ARC', 'ZPN', 'ZEA', 'AIR', 'CYP',
              'CEA', 'CAR', 'MER', 'COP', 'COE',
              'COD', 'COO', 'SFL', 'PAR', 'MOL',
              'AIT', 'BON', 'PCO', 'TSC', 'CSC',
              'QSC', 'HPX', 'XPH']

DEFAULT_SHAPE_OUT = (512, 512)
VALID_GALACTIC = ['galactic', 'g']
VALID_EQUATORIAL = ['celestial2000', 'equatorial', 'eq', 'c', 'q', 'fk4', 'fk5', 'icrs']

logging.basicConfig(format='%(asctime)s -- %(levelname)s: %(message)s', level=logging.DEBUG)

__all__ = ['build_wcs', 'build_wcs_cube', 'build_wcs_2pts',
           'build_ctype', 'equiv_celestial', 'rot_frame',
           'build_wcs_profile']


def equiv_celestial(frame):
    """Return an equivalent ~astropy.coordfinates.builtin_frames

    Notes
    -----
    We do not care of the differences between ICRS/FK4/FK5
    """

    frame = frame.lower()
    if frame in VALID_GALACTIC:
        frame = Galactic()
    elif frame in VALID_EQUATORIAL:
        frame = ICRS()
    elif frame in ['ecliptic', 'e']:
        raise ValueError("Ecliptic coordinate frame not yet supported by astropy")
    return frame


def build_ctype(coordsys, proj_type):
    """Build a valid spatial ctype for a wcs header

    Parameters
    ----------
    coordsys : str ('GALATIC', 'EQUATORIAL')
        the coordinate system of the plate
    proj_type: str ('TAN', 'SIN', 'GSL', ...)
        any projection system supported by WCS

    Returns
    -------
    list:
        a list with the 2 corresponding spatial ctype
    """

    coordsys = coordsys.lower()
    proj_type = proj_type.upper()

    if proj_type not in VALID_PROJ:
        raise ValueError('Unsupported projection')

    if coordsys in VALID_GALACTIC:
        axes = ['GLON-', 'GLAT-']
    elif coordsys in VALID_EQUATORIAL:
        axes = ['RA---', 'DEC--']
    else:
        raise ValueError('Unsupported coordsys')

    return [coord + proj_type for coord in axes]


def rot_frame(coord, proj_sys):
    """Retrieve the proper longitude and latitude

    Parameters
    ----------
    coord : :class:`astropy.coordinate.SkyCoord`
       the sky coordinate of the center of the projection
    proj_sys : str ('GALACTIC', 'EQUATORIAL')
        the coordinate system of the plate (from HEALPIX maps....)

    Returns
    -------
    :class:`~astropy.coordinate.SkyCoord`
        rotated frame
    """

    proj_sys = proj_sys.lower()

    if proj_sys in VALID_EQUATORIAL:
        coord = coord.transform_to(ICRS)
    elif proj_sys in VALID_GALACTIC:
        coord = coord.transform_to(Galactic)
    else:
        raise ValueError('Unsuported coordinate system for the projection')

    return coord


# Cyclic reference if trying to move it to decorator.py
def _lonlat(build_wcs_func):
    """Will decorate the build_wcs function to be able to use it with lon/lat, proj_sys instead of an :class:`astropy.coordinate.SkyCoord` object

    Parameters
    ----------
    build_wcs_func : fct
    a build_wcs_func function (with coords as the first argument

    Returns
    -------
    The same function decorated

    Notes
    -----
    To use this decorator
    build_wcs_lonlat = _lonlat(build_wcs)
    or use @_lonlat on the function declaration

    """

    def decorator(*args, **kwargs):
        """Transform a function call from (lon, lat, src_frame,*) to (coord, *)"""
        if len(args) > 1 and not isinstance(args[0], SkyCoord):

            # Otherwise proceed to fiddling the arguments..
            lon, lat = args[0:2]
            src_frame = kwargs.pop('src_frame', 'EQUATORIAL').lower()

            # Checks proper arguments values
            assert isinstance(lon, float) or isinstance(lon, int), 'lon must be a float'
            assert isinstance(lat, float) or isinstance(lat, int), 'latitude must be a float'
            assert src_frame in VALID_GALACTIC or src_frame in VALID_EQUATORIAL, 'src_frame must be a valid frame'

            frame = equiv_celestial(src_frame)
            coord = SkyCoord(lon, lat, frame=frame, unit="deg")
            # Ugly fix to modigy the args
            args = list(args)
            args.insert(2, coord)
            args = tuple(args[2:])
        return build_wcs_func(*args, **kwargs)

    decorator._coord = build_wcs_func
    decorator.__doc__ = update_docstring(build_wcs_func, skip=6, head_docstring="""
    Parameters
    ----------
    coord : :class:`astropy.coordinate.SkyCoord`
        the sky coordinate of the center of the projection

        or

    lon,lat : floats
        the sky coordinates of the center of projection and
    src_frame :  keyword, str, ('GALACTIC', 'EQUATORIAL')
        the coordinate system of the longitude and latitude (default EQUATORIAL)""", foot_docstring="""
    Notes
    -----
    You can access a function using only catalogs with the ._coord() method
    """)
    return decorator


def build_wcs_profile(pixsize=0.01):

    wcs = WCS(naxis=1)

    # CRPIX IS in Fortran convention -> first pixel edge is 0
    wcs.wcs.crpix = [0.5]
    wcs.wcs.cdelt = [pixsize]
    wcs.wcs.crval = [0]

    wcs.wcs.ctype = ["RADIUS"]

    return wcs


@_lonlat
def build_wcs(coord, pixsize=0.01, shape_out=DEFAULT_SHAPE_OUT, proj_sys='EQUATORIAL', proj_type='TAN'):
    """Construct a :class:`~astropy.wcs.WCS` object for a 2D image

    Parameters
    ----------
    coord : :class:`astropy.coordinate.SkyCoord`
        the sky coordinate of the center of the projection
    pixsize : float
        size of the pixel (in degree)
    shape_out : tuple
        shape of the output map  (n_y,n_x)
    proj_sys : str ('GALACTIC', 'EQUATORIAL')
        the coordinate system of the plate (from HEALPIX maps....)
    proj_type : str ('TAN', 'SIN', 'GSL', ...)
        the projection system to use

    Returns
    -------
    WCS: :class:`~astropy.wcs.WCS`
        An corresponding wcs object
    """

    assert isinstance(shape_out, (tuple, list, np.ndarray))
    assert len(shape_out) == 2

    coord = rot_frame(coord, proj_sys)

    proj_type = proj_type.upper()
    if proj_type not in VALID_PROJ:
        raise ValueError('Unvupported projection')

    wcs = WCS(naxis=2)

    # CRPIX IS in Fortran convention
    wcs.wcs.crpix = (np.array(shape_out, dtype=np.float) + 1) / 2
    wcs.wcs.cdelt = np.array([-pixsize, pixsize])
    wcs.wcs.crval = [coord.data.lon.deg, coord.data.lat.deg]

    wcs.wcs.ctype = build_ctype(proj_sys, proj_type)

    return wcs


@_lonlat
def build_wcs_cube(coord, index, pixsize=0.01, shape_out=DEFAULT_SHAPE_OUT, proj_sys='EQUATORIAL', proj_type='TAN'):
    """Construct a :class:`~astropy.wcs.WCS` object for a 3D cube, where the 3rd dimension is an index

    Parameters
    ----------
    coord : :class:`astropy.coordinate.SkyCoord`
        the sky coordinate of the center of the projection
    index : int
        reference index
    pixsize : float
        size of the pixel (in degree)
    shape_out : tuple
        shape of the output map (n_y, n_x)
    proj_sys : str ('GALACTIC', 'EQUATORIAL')
        the coordinate system of the plate (from HEALPIX maps....)
    proj_type : str ('TAN', 'SIN', 'GSL', ...)
        the projection system to use

    Returns
    -------
    WCS: :class:`~astropy.wcs.WCS`
        An corresponding wcs object
    """

    coord = rot_frame(coord, proj_sys)

    proj_type = proj_type.upper()
    if proj_type not in VALID_PROJ:
        raise ValueError('Unvupported projection')

    wcs = WCS(naxis=3)
    wcs.wcs.crpix = np.append((np.array(shape_out, dtype=np.float) + 1) / 2, 1)
    wcs.wcs.cdelt = np.append(np.array([-pixsize, pixsize]), 1)
    wcs.wcs.crval = np.array([coord.data.lon.deg, coord.data.lat.deg, index])

    wcs.wcs.ctype = np.append(build_ctype(proj_sys, proj_type), 'INDEX').tolist()

    return wcs


def relative_pixsize(coords, pixsize, shape_out, relative_pos):
    """Compute relative_pos or pixsize depending on the input

    Parameters
    ----------
    coords : class:`astropy.coordinate.SkyCoord`
        the 2 sky coordinates of the projection, they will be horizontal in the resulting wcs
    pixsize : float
        size of the pixel (in degree) (default: None, use relative_pos and shape_out)
    shape_out : tuple
        shape of the output map  (n_y,n_x)
    relative_pos : tuple
        the relative position of the 2 sources along the x direction [0-1] (will be computed if pixsize is given)

    Returns
    -------
    tuple
        (pixsize, relative_pos)
    """

    assert len(coords) == 2 & len(shape_out) == 2 & len(relative_pos) == 2, "Must have a length of 2"

    if pixsize:
        # Compute relative_pos from pixsize and distance between
        # sources
        ang_distance = coords[0].separation(coords[1])
        pix_distance = ang_distance.deg / pixsize
        relative_pos = pix_distance / shape_out[1]
        # Center it
        relative_pos = 0.5 + np.array([-1., 1]) * relative_pos / 2
    else:
        # Compute pixsize from relative_pos and distance between
        # sources
        pix_distance = np.array(relative_pos) * shape_out[1]
        pix_distance = np.max(pix_distance) - np.min(pix_distance)
        ang_distance = coords[0].separation(coords[1])
        pixsize = ang_distance.deg / pix_distance

    return pixsize, relative_pos


def build_wcs_2pts(coords, pixsize=None, shape_out=DEFAULT_SHAPE_OUT, proj_sys='EQUATORIAL', proj_type='TAN', relative_pos=(2. / 5, 3. / 5)):
    """Construct a :class:`~astropy.wcs.WCS` object for a 2D image

    Parameters
    ----------
    coords : class:`astropy.coordinate.SkyCoord`
        the 2 sky coordinates of the projection, they will be horizontal in the resulting wcs
    pixsize : float
        size of the pixel (in degree) (default: None, use relative_pos and shape_out)
    shape_out : tuple
        shape of the output map  (n_y,n_x)
    coordsys : str ('GALACTIC', 'EQUATORIAL')
        the coordinate system of the plate (from HEALPIX maps....) will be rotated anyway
    proj_type : str ('TAN', 'SIN', 'GSL', ...)
        the projection system to use, the first coordinate will be the projection center
    relative_pos : tuple
        the relative position of the 2 sources along the x direction [0-1] (will be computed if pixsize is given)

    Returns
    -------
    WCS: :class:`~astropy.wcs.WCS`
        An corresponding wcs object

    Notes
    -----

    By default relative_pos is used to place the sources, and the
    pixsize is derived, but if you define pixsize, then the
    relative_pos will be computed and the sources placed at the center
    of the image

    """

    frame = equiv_celestial(proj_sys)

    proj_type = proj_type.upper()
    if proj_type not in VALID_PROJ:
        raise ValueError('Unsupported projection')

    coords = [coord.transform_to(frame) for coord in coords]
    wcs = WCS(naxis=2)
    pixsize, relative_pos = relative_pixsize(coords, pixsize, shape_out, relative_pos)

    # Put the first source on the relative_pos[0]
    wcs.wcs.crpix = np.array([relative_pos[0], 1. / 2], dtype=np.float) * (
        np.array(shape_out, dtype=np.float)[::-1])
    wcs.wcs.crval = [coords[0].data.lon.deg, coords[0].data.lat.deg]

    wcs.wcs.cdelt = np.array([-pixsize, pixsize])
    # Computes the on-sky position angle (East of North) between this SkyCoord and another.
    rot_angle = (coords[0].position_angle(coords[1]) + Angle(90, unit='deg')).wrap_at('180d')

    logging.debug('... rotating frame with %s deg', rot_angle.degree)
    wcs.wcs.pc = [[np.cos(rot_angle.radian), np.sin(-rot_angle.radian)],
                  [np.sin(rot_angle.radian), np.cos(rot_angle.radian)]]

    wcs.wcs.ctype = build_ctype(proj_sys, proj_type)

    return wcs



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# Copyright (c) 2016 IAS / CNRS / Univ. Paris-Sud
# LGPL License - see attached LICENSE file
# Author: Alexandre Beelen <alexandre.beelen@ias.u-psud.fr>

"""Series of full sky visualization function, with proper wcs header"""

from __future__ import print_function, division
from functools import partial, update_wrapper

import numpy as np

from astropy import units as u
from astropy.io.fits import ImageHDU
from astropy.coordinates import SkyCoord

from .hp_helper import build_wcs, hp_to_wcs, equiv_celestial, _hpmap

__all__ = ['view', 'mollview', 'orthview', 'carview', 'merview',
           'coeview', 'bonview', 'pcoview', 'tscview']

# pv matrix for some of the projection
PV = {'COE': [(1, 1, -20), (2, 1, -70)],
      'BON': [(1, 1, 0), (2, 1, 45)], }


@_hpmap
def view(hp_hdu, coord=None, npix=360, proj_sys='GALACTIC', proj_type='TAN', aspect=1.):
    """projection of the full sky

    Parameters
    ----------
    hp_hdu : :class:`~astropy.io.fits.ImageHDU`
        a pseudo ImageHDU with the healpix map and the associated header to be projected
    coord : :class:`astropy.coordinate.SkyCoord`
        the sky coordinate of the center of the projection
    npix : int
        number of pixels in the latitude direction
    proj_sys : str, ('GALACTIC', 'EQUATORIAL')
        the coordinate system of the projection
    proj_type: str ('TAN', 'SIN', 'GSL', ...)
        any projection system supported by WCS
    aspect : float
        the resulting figure aspect ratio 1:aspect_ratio

    Returns
    -------
    :class:`astropy.io.fits.ImageHDU`
        2D images with header
    """

    if not coord:
        coord = SkyCoord(0, 0, unit='deg', frame=equiv_celestial(proj_sys))

    shape = (np.asarray([1., aspect]) * npix).astype(np.int)
    _wcs = build_wcs._coord(coord, 360. / npix, shape, proj_sys=proj_sys, proj_type=proj_type)

    pv = PV.get(proj_type, None)
    if pv:
        _wcs.wcs.set_pv(pv)
    _data = hp_to_wcs._hphdu(hp_hdu, _wcs, shape[::-1])

    return ImageHDU(_data, _wcs.to_header())


mollview = partial(view, proj_type='MOL', aspect=0.5)
update_wrapper(mollview, view)
mollview.__name__ = "mollview"
mollview.__doc__ = "Mollweide " + mollview.__doc__

carview = partial(view, proj_type='CAR', aspect=0.5)
update_wrapper(carview, view)
carview.__name__ = "carview"
carview.__doc__ = "Plate carrée " + carview.__doc__

merview = partial(view, proj_type='MER', aspect=0.5)
update_wrapper(merview, view)
merview.__name__ = "merview"
merview.__doc__ = "Mercator " + merview.__doc__

coeview = partial(view, proj_type='COE')
update_wrapper(coeview, view)
coeview.__name__ = "coeview"
coeview.__doc__ = "Conic Equal Area " + coeview.__doc__

# TODO: Check ratio
bonview = partial(view, proj_type='BON', aspect=4. / 5)
update_wrapper(bonview, view)
bonview.__name__ = "bonview"
bonview.__doc__ = "Bonne's Equal Area " + bonview.__doc__

# TODO: Check ratio
pcoview = partial(view, proj_type='PCO', aspect=7. / 9)
update_wrapper(pcoview, view)
pcoview.__name__ = "pcoview"
pcoview.__doc__ = "Hassler's polyconic " + pcoview.__doc__

# TODO: Check ratio
tscview = partial(view, proj_type='TSC', aspect=3. / 4)
update_wrapper(tscview, view)
tscview.__name__ = "tscview"
tscview.__doc__ = "Tangential spherical cube " + tscview.__doc__


@_hpmap
def orthview(hp_hdu, coord=None, npix=360, proj_sys='GALACTIC'):
    """Slant orthographic projection of the full sky

    Parameters
    ----------
    hp_hdu : `:class:astropy.io.fits.ImageHDU`
        a pseudo ImageHDU with the healpix map and the associated header to be projected
    coord : :class:`astropy.coordinate.SkyCoord`
        the sky coordinate of the center of the projection
    npix : int
        number of pixels in the latitude direction
    proj_sys : str, ('GALACTIC', 'EQUATORIAL')
        the coordinate system of the projection

    Returns
    -------
    :class:`astropy.io.fits.ImageHDU`
        2D images with header
    """
    if not coord:
        coord = SkyCoord(0, 0, unit='deg', frame=equiv_celestial(proj_sys))

    shape = (npix, npix)

    coord_opposite = SkyCoord(
        coord.data.lon + 180 * u.deg, -1 * coord.data.lat, frame=coord.frame)

    wcs1, wcs2 = [build_wcs._coord(coord, 360. / np.pi / (npix - 1), shape, proj_sys=proj_sys, proj_type='SIN') for coord in [coord, coord_opposite]]

    orth_1 = hp_to_wcs._hphdu(hp_hdu, wcs1, shape[::-1])
    orth_2 = hp_to_wcs._hphdu(hp_hdu, wcs2, shape[::-1])

    return (ImageHDU(orth_1, wcs1.to_header()),
            ImageHDU(orth_2, wcs2.to_header()))



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# Copyright (c) 2016 IAS / CNRS / Univ. Paris-Sud
# LGPL License - see attached LICENSE file
# Author: Alexandre Beelen <alexandre.beelen@ias.u-psud.fr>

r"""
 _   _ ____________          _
| | | || ___ \ ___ \        (_)
| |_| || |_/ / |_/ / __ ___  _
|  _  ||  __/|  __/ '__/ _ \| |
| | | || |   | |  | | | (_) | |
\_| |_/\_|   \_|  |_|  \___/| |
                           _/ |
                          |__/
===============================
HealPix Projection
"""

__version__ = '0.8.1'

from .wcs_helper import *
from .hp_helper import *
from .cutsky import *
from .visu import *



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# Copyright (c) 2016 IAS / CNRS / Univ. Paris-Sud
# LGPL License - see attached LICENSE file
# Author: Alexandre Beelen <alexandre.beelen@ias.u-psud.fr>

try:
    from setuptools import setup
except ImportError:
    from distutils.core import setup

from os import path

here = path.abspath(path.dirname(__file__))

# Get the long description from the README file
with open(path.join(here, 'README.rst')) as f:
    long_description = f.read()


def find_version(filepath):
    """
    Find project version in a given file

    The syntax for the file version need to be in the form
    __version__ = 'a.b.c'
    which follows the semantic versioning http://semver.org/
    * a : major version
    * b : minor version
    * c : patch version

    Parameters
    ----------
    filepath: str
        Path to the file containing a version number

    Returns
    -------
    version: str
        The program version in the form 'a.b.c' as described above

    """
    with open(filepath) as pfile:
        for line in pfile.readlines():
            if line.startswith('__version__'):
                version = line.strip()[-6:-1]
    return version


opts = dict(name="hpproj",
            author='Alexandre Beelen, Marian Douspis',
            author_email='alexandre.beelen@ias.u-psud.fr',
            description='Projection of Healpix maps onto a planar grid',
            long_description=long_description,
            url='https://git.ias.u-psud.fr/abeelen/hpproj',
            download_url='https://git.ias.u-psud.fr/abeelen/hpproj/repository/archive.tar.gz?' + find_version('hpproj/__init__.py'),
            license='LGPL-3.0+',
            classifiers=['Programming Language :: Python :: 2.7',
                         'Programming Language :: Python :: 3.5',
                         'Topic :: Scientific/Engineering :: Astronomy',
                         'Intended Audience :: Science/Research',
                         'License :: OSI Approved :: GNU Lesser General Public License v3 or later (LGPLv3+)'],
            version=find_version('hpproj/__init__.py'),
            packages=['hpproj'],
            package_dir={'hpproj': 'hpproj'},
            entry_points={
                'console_scripts': [
                    'cutsky = hpproj.cutsky:main']},

            setup_requires=['pytest-runner'],
            tests_require=['pytest', 'pytest-mpl', 'scipy'],

            install_requires=[
                'numpy>=1.11',
                'matplotlib>=2.0',
                'astropy>=1.2',
                'healpy>=1.9',
                'photutils>=0.2',
            ],
            )


if __name__ == '__main__':
    setup(**opts)
